package dynclass;

import junit.framework.TestCase;

import java.io.ByteArrayOutputStream;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;

public class RTCompilerTest extends TestCase {

    private class TestLoader extends ClassLoader {

        TestLoader(ClassLoader parent) {

            super(parent);
        }

        Class createClass(String name, byte[] bytes) throws ClassFormatError {

            return defineClass(name, bytes, 0, bytes.length);
        }
    }

    public RTCompilerTest(String name) {

        super(name);
    }

    private Class loadClass(String name, byte[] bytes) throws Exception {

        TestLoader tl = new TestLoader(Thread.currentThread().getContextClassLoader());
        return tl.createClass(name, bytes);
    }

    public void testSubclass() throws Exception {

        Class superClass = java.util.Vector.class;
        RTCompiler rtc = new RTCompiler("dynclass/TestClass",
                                        superClass,
                                        new Class[] { });

        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        rtc.writeClass(baos);
        
        byte[] bytes = baos.toByteArray();
        Class cls = loadClass("dynclass/TestClass", bytes);

        Field[] fields = cls.getDeclaredFields();
        assertEquals(0, fields.length);

        assertEquals(cls.getSuperclass(), superClass);
    }

    public void testInterfaces() throws Exception {

        RTCompiler rtc = new RTCompiler("dynclass/TestClass",
                                        Object.class,
                                        new Class[] { 
                                            Serializable.class,
                                            Cloneable.class
                                        });

        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        rtc.writeClass(baos);
        Class cls = loadClass("dynclass/TestClass", baos.toByteArray());

        Class[] infs = cls.getInterfaces();
        assertEquals(2, infs.length);
        assertEquals(Serializable.class, infs[0]);
        assertEquals(Cloneable.class, infs[1]);
    }

    public void testDeclaredFields() throws Exception {

        Class fieldType1 = Integer.TYPE;
        Class fieldType2 = java.util.Set.class;
        Class fieldType3 = (new String[3][]).getClass();

        RTCompiler rtc = new RTCompiler("dynclass/TestClass",
                                        Object.class,
                                        new Class[] { 
                                            Serializable.class,
                                            Cloneable.class
                                        });
        rtc.addField("intField", fieldType1);
        rtc.addField("setField", fieldType2);
        rtc.addField("arrayField", fieldType3);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        rtc.writeClass(baos);
        Class cls = loadClass("dynclass/TestClass", baos.toByteArray());

        Field[] fields = cls.getDeclaredFields();
        assertEquals(3, fields.length);
        assertEquals("intField", fields[0].getName()); 
        assertEquals("setField", fields[1].getName()); 
        assertEquals("arrayField", fields[2].getName()); 
        assertEquals(fieldType1, fields[0].getType()); 
        assertEquals(fieldType2, fields[1].getType());
        assertEquals(fieldType3, fields[2].getType());
        assertTrue(Modifier.isPrivate(fields[0].getModifiers()));
        assertTrue(Modifier.isPrivate(fields[1].getModifiers()));
        assertTrue(Modifier.isPrivate(fields[2].getModifiers()));
    }

    public void testAddConstructor() throws Exception {

        RTCompiler rtc = new RTCompiler("dynclass/TestClass",
                                        Object.class,
                                        new Class[] { });

        ConstantPool cp = rtc.getConstantPool();
        short initIndex = cp.methodRefConst("java/lang/Object", "<init>", "()V");
        //fail("method index: " + initIndex + "; cp:" + cp);
        Class[] none = new Class[0];
        byte[] bc = {
            rtc.ALOAD_0,
            rtc.INVOKESPECIAL,
            (byte) (initIndex >> 8),
            (byte) (initIndex % 256),
            rtc.RETURN,
        };
        rtc.addMethod("<init>", Void.TYPE, none, none, (short)1, (short)1, bc);
        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        rtc.writeClass(baos);
        Class cls = loadClass("dynclass/TestClass", baos.toByteArray());

        Constructor ctor = cls.getConstructor(none);
        Object instance = ctor.newInstance(null);
        assertEquals(cls, instance.getClass());
    }
}
